;file lib_i2cMaster.inc

_lib_i2cm_var		macro	nr, prefix
			org	data
expand
vp??nr??_bank		equ	($ & $F8) | nr
lib_i2cm_sda_pin	=	$+0	; pin used for SDA (highnib = port, lownib = pin)
lib_i2cm_scl_pin	=	$+1	; pin used for SCL (highnib = port, lownib = pin)
lib_i2cm_state		=	$+2	; This indicates the state that the i2cm master is currently in.
lib_i2cm_sub_state	=	$+3	; This indicates the substate that the I2C master is currently in.
lib_i2cm_bit_count	=	$+4	; Indicates the number of bits left to process in read/write
lib_i2cm_byte		=	$+5	; The byte currently being written/read by the I2C master
lib_i2cm_address	=	$+6	; The address to read/write to.
lib_i2cm_flags		=	$+7
  lib_i2cm_nack		=	lib_i2cm_flags.4	; This bit is set if the I2C master has received a NACK from the slave
  lib_i2cm_data_needed	=	lib_i2cm_flags.6	; set when master is ready to send data
  lib_i2cm_rx_flag	=	lib_i2cm_flags.7	; set when master has received a byte
prefix??_??bank		equ	($ & $F0) | nr
prefix??_??sda_pin	ds	1	; pin used for SDA (highnib = port, lownib = pin)
prefix??_??scl_pin	ds	1	; pin used for SCL (highnib = port, lownib = pin)
prefix??_??state	ds	1	; This indicates the state that the i2cm master is currently in.
prefix??_??sub_state	ds	1	; This indicates the substate that the I2C master is currently in.
prefix??_??bit_count	ds	1	; Indicates the number of bits left to process in read/write
prefix??_??byte		ds	1	; The byte currently being written/read by the I2C master
prefix??_??address	ds	1	; The address to read/write to.
prefix??_??flags	ds	1
  prefix??_??nack	equ	prefix??_??flags.4	; This bit is set if the I2C master has received a NACK from the slave
  prefix??_??data_needed	equ	prefix??_??flags.6	; set when master is ready to send data
  prefix??_??rx_flag	equ	prefix??_??flags.7	; set when master has received a byte
noexpand
data			=	$
endm

lib_i2cMaster		macro	nr, prefix, sdapin, sclpin, thread
if nr >= maxvp
  error "vp number too high"
endif
ifndef vp_i2cMaster
vp_i2cMaster
endif
ifndef vp_dirbuf
vp_dirbuf
endif
vp??nr??_thread	= thread
_lib_i2cm_var nr,prefix
		org	code
expand
	;*********************************************************************************
	; Function: i2cm_init
	;	This subroutine should be called on startup.  It initializes the registers
	;	which are critical to the operation of the i2cm state machine.
	;*********************************************************************************
vp??nr??_init
		mov	fsr,#vp??nr??_bank
		setb	lib_i2cm_data_needed
		mov	lib_i2cm_sda_pin,#sdapin
		mov	lib_i2cm_scl_pin,#sclpin
noexpand
code		=	$
expand
		org	vp??nr??_isr
		jmp	@i2cm_isr
noexpand
endm

_lib_i2cMaster_lowsize	equ	$049
_lib_i2cMaster_lowhalf	macro
			org	code
	begin	=	$
expand
;*****************************************************************************************
; I2C Master Interrupt Service Routines
;*****************************************************************************************
; I2C Virtual Peripheral...

i2cm_clrbSDA	mov	w,lib_i2cm_sda_pin
		jmp	@clrb_dirbuf
i2cm_setbSDA	mov	w,lib_i2cm_sda_pin
		jmp	@setb_dirbuf
i2cm_clrbSCL	mov	w,lib_i2cm_scl_pin
		jmp	@clrb_dirbuf
i2cm_setbSCL	mov	w,lib_i2cm_scl_pin
		jmp	@setb_dirbuf
i2cm_do_i2cm_start_write	jmp	_do_i2cm_start_write
i2cm_do_i2cm_start_read		jmp	_do_i2cm_start_read
i2cm_do_i2cm_write		jmp	_do_i2cm_write
i2cm_do_i2cm_read		jmp	_do_i2cm_read
i2cm_do_i2cm_read_last		jmp	_do_i2cm_read_last
i2cm_do_i2cm_stop		jmp	_do_i2cm_stop

	;*********************************************************************************
	; Function: i2cm_ISR
	; I2C Master Interrupt-Driven State Machine
	; -------------------------------------------
	;	This is the I2C Master Interrupt Service Routine.  It is an interrupt-
	;	driven state machine which allows all of the actions of the I2C Master
	;	controller to be carried out, virtual peripheral style, with virtually
	;	no interaction from the mainline program.  
	;  i2cm_idle
	; 	This is the state that the I2C Master is usually in when it is not in use.
	; 	It just ensures that the I2C_port_buf SCL and SDA are both set high
	;  i2cm_start
	; 	When any mainline program wants to use the I2C master it puts the master into
	; 	start mode.  This mode creates a start condition on the I2C bus.  A start 
	;	condition is created when SDA goes from high to low while SCL stays high.
	;  i2cm_start_write
	;	This state performs some pre-processing which allows the i2cm_write state to
	;	do its work.  It sets up the bit count, gets the next piece of data from the
	;	buffer and prepares to send it.
	;  i2cm_write
	;	This state writes the data in i2cm_byte to the I2C bus
	;  i2cm_get_ack
	;	This state gets an ACK from the slave device.  If no ACK is received, the I2C
	;	Master state machine puts a stop condition on the bus and the i2cm_flags 
	;	register is loaded to indicate that a NACK has occurred.
	;  i2cm_write_repeat
	;	This state determines, after one byte of data is sent, whether or not
	; 	there is another byte to be sent.  If so, this state goes back to i2cm_start_write
	;	and sends the next byte.
	;  i2cm_stop
	;	This state puts a stop condition on the I2C bus and resets the state machine back
	;	to its idle state.  A stop condition is when SDA goes from low to high while SCL
	;	is high.
	;  i2cm_start_read
	;	This state simply loads the contents of the i2cm_address register into the i2cm_byte
	;	register and sets up the i2cm_write state to output the address of the slave to read.
	;  i2cm_read_data	
	; 	This state prepares the i2cm read routine so it can read from the slave device.  It
	;	initializes the bit count, etc.
	;  i2cm_read
	; 	This state read 8 bits of data from the slave device.
	;  i2cm_send_ack
	; 	This state sends an ack if there is data left to write, and a NACK if there is
	;	no data left to write.
	;*********************************************************************************
i2cm_isr
		mov	w,lib_i2cm_state		;1
		add	PC,w			;3 ;Add the state to the program counter
						   ;and go to the state in the jump table.
		;*************************************************************************
		; States for I2C Master while idle
		;*************************************************************************
	i2cm_idle_loc		=	$
		jmp	i2cm_idle		;3 ;If I2C_state = 0, I2C is idle

		;*************************************************************************
		; States for I2C Master while writing
		;*************************************************************************
	i2cm_write_loc	=	$
		jmp	i2cm_start		;3 ;Make SDA go low while SCL is high
		jmp	i2cm_start_write	;3 ;Load a byte from the buffer and prepare to send. 
	i2cm_write_again = $
		jmp	i2cm_write		;3 ;Write it.
		jmp	i2cm_get_ack		;3 ;Get an ACK signal
		jmp	i2cm_write_repeat	;3 ;Check to see if we have finished sending 
						   ;(if buffer read index = buffer write index)
	i2cm_stop_loc	=	$
		jmp	i2cm_stop		;3 ;If write_repeat determines we're finished,
						   ;then send stop

		;*************************************************************************
		; States for I2C Master while reading
		;*************************************************************************
	i2cm_read_loc	=	$		   ;Load this state into state machine if we
						   ;are starting to read
		jmp	i2cm_start		;3 ;
		jmp	i2cm_start_read		;3 ;Write address used for reading
		jmp	i2cm_write 		;3 ;(writing address)
		jmp	i2cm_get_ack		;3 ;
	i2cm_read_rpt_loc =	$
		jmp	i2cm_data_ready		;3 ;wait until mainline code has grabbed data.
		jmp	i2cm_read_data		;3 ;
		jmp	i2cm_read		;3 ;Keep doing this until done
		jmp	i2cm_send_ack		;3 ;
		jmp	i2cm_stop		;3 ;
						
	;*********************************************************************************
	; State: i2cm_start
	; 	When any mainline program wants to use the I2C master it puts the master into
	; 	start mode.  This mode creates a start condition on the I2C bus.  A start 
	;	condition is created when SDA goes from high to low while SCL stays high.
	;*********************************************************************************
i2cm_start	mov	w,lib_i2cm_sub_state	;1
		add	pc,w			;3
		jmp	_i2cm_start:state1			;3
		jmp	_i2cm_start:state2			;3
		jmp	_i2cm_start:state3			;3
		jmp	_i2cm_start:state4			;3
		jmp	_i2cm_start:state5			;3

	;*********************************************************************************
	; State: i2cm_write
	;	This state writes the data in i2cm_byte to the I2C bus
	;*********************************************************************************
i2cm_write	mov	w,lib_i2cm_sub_state	;1
		add	pc,w			;3
		jmp	_i2cm_write:state1			;3
		jmp	_i2cm_write:state2			;3
		jmp	_i2cm_write:state3			;3

	;*********************************************************************************
	; State: i2cm_get_ack
	;	This state gets an ACK from the slave device.  If no ACK is received, the I2C
	;	Master state machine puts a stop condition on the bus and the i2cm_flags 
	;	register is loaded to indicate that a NACK has occurred.
	;*********************************************************************************
i2cm_get_ack	mov	w,lib_i2cm_sub_state	;1
		add	pc,w			;3
		jmp	_i2cm_get_ack:state1			;3
		jmp	_i2cm_get_ack:state2			;3
		jmp	_i2cm_get_ack:state3			;3
		jmp	_i2cm_get_ack:state4			;3

	;*********************************************************************************
	; State: i2cm_stop
	;	This state puts a stop condition on the I2C bus and resets the state machine back
	;	to its idle state.  A stop condition is when SDA goes from low to high while SCL
	;	is high.
	;*********************************************************************************
i2cm_stop	mov	w,lib_i2cm_sub_state	;1
		add	pc,w			;3
		jmp	_i2cm_stop:state1			;3
		jmp	_i2cm_stop:state2			;3
		jmp	_i2cm_stop:state3			;3

	;*********************************************************************************
	; Function: i2cm_read_data	
	; 	This function prepares the i2cm read routine so it can read from the slave 
	;	device. It initializes the bit count, etc.
	;*********************************************************************************
i2cm_read_data
		mov	w,#8
		mov	lib_i2cm_bit_count,w
		inc	lib_i2cm_state
		;*************************************************************************
		; Read 8 bits of data from the slave device.
		;*************************************************************************
i2cm_read	mov	w,lib_i2cm_sub_state
		add	pc,w
		jmp	_i2cm_read:state1
		jmp	_i2cm_read:state2
		jmp	_i2cm_read:state3

	;*********************************************************************************
	; Function: i2cm_send_ack
	; 	This function sends an ack to read more bytes
	;*********************************************************************************
i2cm_send_ack
		mov	w,lib_i2cm_sub_state
		add	pc,w
		jmp	_i2cm_send_ack:state1
		jmp	_i2cm_send_ack:state2
		jmp	_i2cm_send_ack:state3
		jmp	_i2cm_send_ack:state4	
noexpand
code			=	$
if _lib_i2cMaster_lowsize > code - begin
  error P2W "_i2cMaster_lowsize too large"
endif
if _lib_i2cMaster_lowsize < code - begin
  error "_i2cMaster_lowsize too small"
endif
endm

_lib_i2cMaster_highsize	equ	$09C
_lib_i2cMaster_highhalf	macro
		org	code
	begin	=	$
expand
_i2cm_start
:state1	 	call	i2cm_setbSCL
		call	i2cm_setbSDA
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 9 + 7 = 16

;	--- WAITING FOR IDLE SCL ---
:state2
:m1		mov	w,lib_i2cm_scl_pin
		call	@get_pin
		snz
		clr	lib_i2cm_sub_state		;1 
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 6/8 + 7 = 13/15

:state3		call	i2cm_clrbSDA
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 5 + 7 = 12

:state4		call	i2cm_clrbSCL
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 5 + 7 = 12

:state5		call	i2cm_setbSDA
		clr	lib_i2cm_sub_state		;1
		inc	lib_i2cm_state		;1
		retp				;3 = 6 + 7 = 13

_i2cm_write
:state1		call	i2cm_setbSDA
		rl	lib_i2cm_byte		;1
		sc				;1
		call	i2cm_clrbSDA
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 15 + 7 = 22

:state2		call	i2cm_setbSCL
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 12 + 7 = 19

:state3		call	i2cm_clrbSCL
		dec	lib_i2cm_bit_count		;1
		snz				;1
		jmp	:done			;3
		clr	lib_i2cm_sub_state		;1
		retp				;3 = 14 + 7 = 21

:done		clr	lib_i2cm_sub_state		;1
		inc	lib_i2cm_state		;1
		retp				;3 = 18 + 7 = 25

_i2cm_get_ack
:state1		call	i2cm_setbSDA
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 13 + 7 = 20

:state2		call	i2cm_setbSCL
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 12 + 7 = 19

:state3
:m1		mov	w,lib_i2cm_sda_pin
		call	@get_pin
		jnz	:nack		;1
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 13 + 7 = 20 ;if no ack,
:nack		mov	w,#(i2cm_stop_loc-i2cm_idle_loc);1
		mov	lib_i2cm_state,w		;1
		clr	lib_i2cm_sub_state		;1 ;send a stop and indicate that this didn't work.
		setb	lib_i2cm_nack		;1 ;set i2cm_NACK to show that this did not go through.
		retp				;3 = 17 + 7 = 24

:state4		call	i2cm_clrbSCL
		clr	lib_i2cm_sub_state		;1
		clrb	lib_i2cm_nack
		inc	lib_i2cm_state		;1 ;move on to next state
		retp				;3 = 13 + 7 = 20

_i2cm_stop
:state1		call	i2cm_clrbSDA
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 13 + 7 = 20

:state2		call	i2cm_setbSCL
		inc	lib_i2cm_sub_state		;1
		retp				;3 = 13 + 7 = 20

:state3		call	i2cm_setbSDA
		clr	lib_i2cm_sub_state		;1 ;put the state machine in Idle
		clr	lib_i2cm_state		;1
		retp				;3 = 16 + 7 = 23 

_i2cm_read
:state1		call	i2cm_setbSCL
		inc	lib_i2cm_sub_state
		retp

:state2
		mov	w,lib_i2cm_sda_pin
		call	@get_pin
		stc
:m1		snz
		clc
:m_done		rl	lib_i2cm_byte
		inc	lib_i2cm_sub_state
		retp

:state3		call	i2cm_clrbSCL
		clr	lib_i2cm_sub_state
		dec	lib_i2cm_bit_count
		sz	
		retp
		setb	lib_i2cm_rx_flag
		inc	lib_i2cm_state
		retp

_i2cm_send_ack
:state1		call	i2cm_clrbSDA
		inc	lib_i2cm_sub_state
		retp

:state2		call	i2cm_setbSCL		;Pulse the clock
		inc	lib_i2cm_sub_state
		retp

:state3		call	i2cm_clrbSCL
		inc	lib_i2cm_sub_state
		retp

:state4		call	i2cm_setbSDA
		clr	lib_i2cm_sub_state
		mov	w,#(i2cm_read_rpt_loc - i2cm_idle_loc) ;back to read 
		mov	lib_i2cm_state,w
		retp

	;*********************************************************************************
	; State: i2cm_idle
	; 	This is the state that the I2C Master is usually in when it is not in use.
	; 	It just ensures that the I2C_port_buf SCL and SDA are both set high
	;*********************************************************************************
i2cm_idle	call	i2cm_setbSDA
		call	i2cm_setbSCL
		setb	lib_i2cm_data_needed
		retp

	;*********************************************************************************
	; State: i2cm_start_write	
	;	This state sets up the bit count and prepares to send the byte in i2cm_byte
	;*********************************************************************************
i2cm_start_write
		mov	w,#8			;1
		mov	lib_i2cm_bit_count,w	;1
		inc	lib_i2cm_state		;1
		retp				;3 = 12 + 7 = 19

	;*********************************************************************************
	; State: i2cm_data_ready
	;	Wait for mainline code to grab data
	;*********************************************************************************
i2cm_data_ready
		snb	lib_i2cm_rx_flag
		retp
		inc	lib_i2cm_state
		retp

	;*********************************************************************************
	; State: i2cm_write_repeat
	;	Wait for mainline code to put another byte in i2cm_byte and raising i2cm_doing_write
	; 	Then this state goes back to i2cm_start_write and sends the next byte.
	;*********************************************************************************
i2cm_write_repeat
		snb	lib_i2cm_data_needed
		retp
		dec	lib_i2cm_state		;1 ;back to get_ack
		dec	lib_i2cm_state		;1 ;back to write
		dec	lib_i2cm_state		;1 ;back to write_data
		retp				;3 = 16 + 7 =23

	;*********************************************************************************
	; Function: i2cm_start_read
	;	This state simply loads the contents of the i2cm_address register into the 
	;	i2cm_byte register and sets up the i2cm_write state to output the address 
	;	of the slave to read.
	;*********************************************************************************
i2cm_start_read
		mov	w,#8
		mov	lib_i2cm_bit_count,w
		inc	lib_i2cm_state
		retp

	;*********************************************************************************
	; Function: i2cm_send_nack
	; 	This function sends a nack to inform slave no more data will be read from it
	;*********************************************************************************
i2cm_send_nack
		call	i2cm_setbSDA
		clr	lib_i2cm_sub_state
		inc	lib_i2cm_state		;no more bytes to read, send a stop
		retp

;*****************************************************************************************
; End of I2C Master state machine
;*****************************************************************************************
;*****************************************************************************************
; I2C Master Subroutines:  Mainline Access routines (call from the mainline)
;*****************************************************************************************

	;*********************************************************************************
	; Function: do_i2cm_start_write
	;	Wait until i2cm_data_needed is high
	;	Load w with slave address
	;	Issue call
	;*********************************************************************************
_do_i2cm_start_write
		mov	lib_i2cm_address,w
		mov	lib_i2cm_byte,w
		clrb	lib_i2cm_byte.0
		mov	lib_i2cm_state,#(i2cm_write_loc-i2cm_idle_loc)
		clrb	lib_i2cm_data_needed
		retp

	;*********************************************************************************
	; Function: do_i2cm_start_read
	;	Wait until i2cm_data_needed is high
	;	Load w with slave address
	;	Issue call
	;*********************************************************************************
_do_i2cm_start_read
		mov	lib_i2cm_address,w
		mov	lib_i2cm_byte,w
		setb	lib_i2cm_byte.0
		mov	lib_i2cm_state,#(i2cm_read_loc-i2cm_idle_loc)
		clrb	lib_i2cm_data_needed
		retp

	;*********************************************************************************
	; Function: do_i2cm_write
	;	Wait until i2cm_data_needed is high
	;	Load w with data
	;	Issue call
	;*********************************************************************************
_do_i2cm_write
		mov	lib_i2cm_byte,w
		clrb	lib_i2cm_data_needed
		retp

	;*********************************************************************************
	; Function: do_i2cm_read
	;	Wait until i2cm_rx_flag is high
	;	Issue call
	;	data returned in w
	;*********************************************************************************
_do_i2cm_read
		mov	w,lib_i2cm_byte
		clrb	lib_i2cm_rx_flag
		retp

_do_i2cm_read_last
	;*********************************************************************************
	; Function: do_i2cm_stop
	;	Wait until i2cm_data_needed is high
	;	Issue call
	;*********************************************************************************
_do_i2cm_stop
		mov	lib_i2cm_state,#(i2cm_stop_loc-i2cm_idle_loc)
		clrb	lib_i2cm_data_needed
		retp
noexpand
code			=	$
if _lib_i2cMaster_highsize > code - begin
  error P2W "_i2cMaster_highsize too large"
endif
if _lib_i2cMaster_highsize < code - begin
  error "_i2cMaster_highsize too small"
endif
endm

;end of file lib_i2cMaster.inc
