;file lib_i2cSlave.inc

_lib_i2cs_var		macro	nr, prefix
			org	data
expand
vp??nr??_bank		equ	($ & $F8) | nr
lib_i2cs_sda_pin		=	$+0	; pin used for SDA (highnib = port, lownib = pin)
lib_i2cs_scl_pin		=	$+1	; pin used for SCL (highnib = port, lownib = pin)
lib_i2cs_state			=	$+2	; This indicates the state that the I2C slave is currently in.
lib_i2cs_sub_state		=	$+3	; This indicates the substate that the I2C slave is currently in.
lib_i2cs_bit_count		=	$+4	; Indicates the number of bits left to process in read/write
lib_i2cs_byte			=	$+5	; The byte currently being written/read by the I2C master
lib_i2cs_address		=	$+6	; This byte holds the address which the I2C slave will respond to.
lib_i2cs_flags			=	$+7
  lib_i2cs_pres_sda		=	lib_i2cs_flags.0 ; The present state of the SCL line
  lib_i2cs_pres_scl		=	lib_i2cs_flags.1 ; The present state of the SDA line
  lib_i2cs_past_sda		=	lib_i2cs_flags.2 ; The state of the SCL line on last interrupt
  lib_i2cs_past_scl		=	lib_i2cs_flags.3 ; The state of the SDA line on last interrupt
  lib_i2cs_stop_found		=	lib_i2cs_flags.4
  lib_i2cs_written_again	=	lib_i2cs_flags.5
  lib_i2cs_data_needed		=	lib_i2cs_flags.6	; set if master requested more data
  lib_i2cs_rx_flag		=	lib_i2cs_flags.7	; This flag indicates when data has
prefix??_??bank		equ	($ & $F0) | nr
prefix??_??sda_pin	ds	1	; pin used for SDA (highnib = port, lownib = pin)
prefix??_??scl_pin	ds	1	; pin used for SCL (highnib = port, lownib = pin)
prefix??_??state	ds	1	; This indicates the state that the I2C slave is currently in.
prefix??_??sub_state	ds	1	; This indicates the substate that the I2C slave is currently in.
prefix??_??bit_count	ds	1	; Indicates the number of bits left to process in read/write
prefix??_??byte	ds	1	; The byte currently being written/read by the I2C master
prefix??_??address	ds	1	; This byte holds the address which the I2C slave will respond to.
prefix??_??flags	ds	1
  prefix??_??pres_scl		equ	prefix??_??flags.0 ; The present state of the SCL line
  prefix??_??pres_sda		equ	prefix??_??flags.1 ; The present state of the SDA line
  prefix??_??past_scl		equ	prefix??_??flags.2 ; The state of the SCL line on last interrupt
  prefix??_??past_sda		equ	prefix??_??flags.3 ; The state of the SDA line on last interrupt
  prefix??_??stop_found	equ	prefix??_??flags.4
  prefix??_??written_again	equ	prefix??_??flags.5
  prefix??_??data_needed	equ	prefix??_??flags.6	; set if master requested more data
  prefix??_??rx_flag		equ	prefix??_??flags.7	; This flag indicates when data has
noexpand
data			=	$
endm

lib_i2cSlave		macro	nr, prefix, sdapin, sclpin, thread
if nr >= maxvp
  error "vp number too high"
endif
ifndef vp_i2cSlave
vp_i2cSlave
endif
ifndef vp_dirbuf
vp_dirbuf
endif
vp??nr??_thread	= thread
_lib_i2cs_var nr,prefix
		org	code
expand
	;*********************************************************************************
	; Function: i2cs_init
	;	This subroutine should be called on initialization of the program.  It
	;	initializes the variables which are critical to the operation of the 
	;	i2cs slave state machine.
	;*********************************************************************************
vp??nr??_init
		mov	fsr,#vp??nr??_bank
		mov	w,#$0f			;Let part know last states have all been highs
		mov	lib_i2cs_flags,w
		mov	lib_i2cs_sda_pin,#sdapin
		mov	lib_i2cs_scl_pin,#sclpin
noexpand
code		=	$
expand
		org	vp??nr??_isr
		jmp	@i2cs_isr
noexpand
endm

_lib_i2cSlave_lowsize	equ	$052
_lib_i2cSlave_lowhalf	macro
sda		equ	0
scl		equ	1
			org	code
	begin	=	$
expand
;*****************************************************************************************
; I2C Slave Interrupt Service Routines
;*****************************************************************************************

i2cs_clrbSDA	mov	w,lib_i2cs_sda_pin
		jmp	@clrb_dirbuf
i2cs_setbSDA	mov	w,lib_i2cs_sda_pin
		jmp	@setb_dirbuf
i2cs_clrbSCL	mov	w,lib_i2cs_scl_pin
		jmp	@clrb_dirbuf
i2cs_setbSCL	mov	w,lib_i2cs_scl_pin
		jmp	@setb_dirbuf

	;*********************************************************************************
	; Function: i2cs_ISR
	; I2C Slave Interrupt-Driven State Machine
	;	    ----------------------------------------
	;
	;	This is the main interrupt service routine for the I2C slave interrupt
	;	service routine.  It is a state machine, allowing semi-straightline 
	;	code to operate on successive calls.  The current state of this state
	;	machine is stored in the registers i2cs_state and i2cs_sub_state.  These
	;	are the discrete states of this state machine...
	;
	;  i2cs_idle
	;	i2cs_idle is the state used when no start condition has occurred
	;	and the device has not been addressed.  The I2C slave will sit in this
	;	state until the state is changed by the independant subroutine 
	;	i2cs_get_start_stop.  
	;
	;  i2cs_wait_for_scl_low
	;	This state simply waits until SCL goes low.  It is entered once
	;	i2cs_get_start_stop encounters a start condition.  Once SCL goes low,
	;	the state is incremented
	;
	;  i2cs_get_address
	;	This state prepares the i2cs_read_byte routine to get 8-bits of data. 
	;	It increments the state to i2cs_read_byte before exiting.
	;
	;  i2cs_read_byte
	;	This state reads a byte of data.  i2cs_bit_count needs to be loaded with
	;	#8 before entering this state, or it won't know to increment to the next
	;	state after 8 bits of data have been loaded.
	;
	;  i2cs_process_address
	;	This state simply performs a quick calculation to figure out if this slave
	;	was just addressed.  If it was not, then it changes the state back to 
	;	i2cs_idle.  If this slave was addressed, then this state prepares to either
	;	read data or write data, depending on the bit 1 of the first byte received.
	;
	;  i2cs_send_ack
	;	This state outputs an ACK pulse, to tell the master that data was received
	;	correctly.  It pulls SDA low while SCL is pulsed high and low by the master.
	;
	;  i2cs_read_data
	;	This state prepares i2cs_read_byte to receive 8 bits of data.  It increments
	;	the state before exiting.
	;
	;  i2cs_make_idle
	;	This state puts the I2C slave back into idle mode.
	;
	;  i2cs_process_data
	;	This state processes a byte of data which was just received.  It moves the 
	;	byte that was just received into the i2cs_data_in register, and sets the 
	;	i2cs_event_flag to indicate an i2cs event and sets the i2cs_rx_flag to 
	;	indicate that a byte of data was received by the slave.
	;
	;  i2cs_send_data
	;	This state prepares the i2cs_write_byte state to send the 8-bits of data
	;	in the i2cs_data_out register.
	;
	;  i2cs_write_byte
	;	This state outputs a byte of data, clocked out by the SCL pin.  It must be
	;	pre-prepared to send out a byte by having the i2cs_byte register loaded with
	;	valid data and having the i2cs_bit_count register loaded with #8.
	;
	;  i2cs_get_ack
	;	This state gets an ACK from the I2C master.  If an ACK is received, this state
	;	will try to send another byte of data from the i2cs_data_out register.  If no
	;	ACK is received, the slave will be put back into its idle state.
	;*********************************************************************************
i2cs_isr
	;*********************************************************************************
	; Function: i2cs_get_start_stop
	; This subroutine checks the i2cs_past_pres register for the presence of a start
	; or stop condition.  It takes the values present in the i2cs_past_pres register
	; and checks them against what would be in the register if a start or stop condition
	; had occured.  This subroutine should be called on every pass of the ISR, because
	; it runs independantly to the rest of the slave code and re-initializes the state
	; machine, regardless of its last state, on the receipt of these conditions.
	;*********************************************************************************
i2cs_get_start_stop
		mov	vp_portpin,lib_i2cs_flags
		mov	w,lib_i2cs_scl_pin
		call	@get_pin
		clc				;1
		sz
		stc				;1
		rl	vp_portpin		;1 shiftin current scl state
		mov	w,lib_i2cs_sda_pin
		call	@get_pin
		clc				;1
		sz
		stc				;1
		rl	vp_portpin		;1 shiftin current sda state
		and	lib_i2cs_flags,#$F0	;1
		mov	w,vp_portpin
		and	w,#$0F
		or	lib_i2cs_flags,w	;1 save sda and scl states
		xor	w,#%00001110		;1 ;Compare with a start condition.
		snz				;1
		jmp	:start			;3
		xor	w,#%00000101		;1 ;Compare XOR of start and stop with
		sz				;1 ;what would be present if a stop was present.
		jmp	_i2cs_isr

:stop		clr	lib_i2cs_state		;1 ;If stop, then make the I2C slave idle
		clr	lib_i2cs_sub_state		;1
		setb	lib_i2cs_stop_found		;1
		jmp	_i2cs_isr

:start		mov	w,#1			;1 ;If start, then start receiving the address.
		mov	lib_i2cs_state,w		;1

_i2cs_isr	mov	w,lib_i2cs_state	;1 
		add	pc,w		;3 ;Add the state to the program counter
					   ;and go to the state in the jump table.

		;*************************************************************************
		; States for I2C slave while idle
		;*************************************************************************
	i2cs_idle_loc		=	$
		jmp	i2cs_idle		;3 ;If I2C_state = 0, I2C is idle
		jmp	i2cs_wait_for_scl_low	;3 ;Wait for SCL line to go low after start condition
		jmp	i2cs_get_address	;3 ;Get the address
		jmp	i2cs_read_byte		;3 
		jmp	i2cs_process_address	;3 ;Check to see if this device was addressed, and set
						   ;up for read or write if it was.
		;*************************************************************************
		; States for I2C slave being read
		;*************************************************************************
	i2cs_being_read		=	$	; Go to this state if the slave is being read.
		jmp	i2cs_send_ack		;3 
	i2cs_being_read_again	=	$
		jmp	i2cs_request_for_data	;3 ;tell mainline code to supply data.
		jmp	i2cs_send_data		;3 ;Set up to send out a byte.
		jmp	i2cs_write_byte		;3 ;Send the byte, clocked by SCL
		jmp	i2cs_get_ack		;3 ;If we receive an ACK, then we are being read again.
		jmp	i2cs_waiting_for_data	;3 ;If there is no valid data, wait for it.
	
		;*************************************************************************
		; States for I2C slave being written
		;*************************************************************************
	i2cs_being_written	=	$	; Go to this state if the slave is being written
		jmp	i2cs_send_ack		;3 ;Send an ACK
	i2cs_being_written_again	=	$
		jmp	i2cs_read_data		;3 
		jmp	i2cs_read_byte		;3 
		jmp	i2cs_process_data	;3 
		jmp	i2cs_send_ack		;3 
		jmp	i2cs_read_again		;3 

	;*********************************************************************************
	; State: i2cs_read_byte
	;	This state reads a byte of data.  i2cs_bit_count needs to be loaded with
	;	#8 before entering this state, or it won't know to increment to the next
	;	state after 8 bits of data have been loaded.
	;*********************************************************************************
i2cs_read_byte
		mov	w,lib_i2cs_sub_state	;1
		add	pc,w			;3
		jmp	_i2cs_read_byte:state1		;1
		jmp	_i2cs_read_byte:state2			;3

	;*********************************************************************************
	; State: i2cs_send_ack
	;	This state outputs an ACK pulse, to tell the master that data was received
	;	correctly.  It pulls SDA low while SCL is pulsed high and low by the master.
	;*********************************************************************************
i2cs_send_ack	
		mov	w,lib_i2cs_sub_state	;1
		add	pc,w			;3
		jmp	_i2cs_send_ack:state1			;3
		jmp	_i2cs_send_ack:state2			;3
		jmp	_i2cs_send_ack:state3			;3

	;*********************************************************************************
	; State: i2cs_write_byte
	;	This state outputs a byte of data, clocked out by the SCL pin.  It must be
	;	pre-prepared to send out a byte by having the i2cs_byte register loaded with
	;	valid data and having the i2cs_bit_count register loaded with #8.
	;*********************************************************************************
i2cs_write_byte
		mov	w,lib_i2cs_sub_state	;1
		add	pc,w			;3
		jmp	_i2cs_write_byte:state1			;3
		jmp	_i2cs_write_byte:state2			;3
		jmp	_i2cs_write_byte:state3			;3

	;*********************************************************************************
	; State: i2cs_get_ack
	;	This state gets an ACK from the I2C master.  If an ACK is received, this state
	;	will try to send another byte of data from the i2cs_data_out register.  If no
	;	ACK is received, the slave will be put back into its idle state.
	;*********************************************************************************
i2cs_get_ack
		mov	w,lib_i2cs_sub_state	;1
		add	pc,w			;3
		jmp	_i2cs_get_ack:state1			;3
		jmp	_i2cs_get_ack:state2			;3
noexpand
code		=	$
if _lib_i2cSlave_lowsize > code - begin
  error P2W "_i2cSlave_lowsize too large"
endif
if _lib_i2cSlave_lowsize < code - begin
  error "_i2cSlave_lowsize too small"
endif
endm

_lib_i2cSlave_highsize	equ	$07A
_lib_i2cSlave_highhalf	macro
		org	code
	begin	=	$
expand
_i2cs_read_byte
:state1						; wait until SCL goes high
		call	i2cs_setbSCL
		sb	lib_i2cs_pres_scl		;1
		retp				;3 = 10 + 7 = 17		
		inc	lib_i2cs_sub_state		;1 ;If SCL is high, rotate in the bit
		clc				;1
		snb	lib_i2cs_pres_sda		;1
		stc				;1
		rl	lib_i2cs_byte		;1
		retp				;3 = 15/17 + 7 = 22/24
:state2						; wait until SCL goes low
		snb	lib_i2cs_pres_scl		;1
		retp				;3 = 11 + 7 = 18
		call	i2cs_clrbSCL
		clr	lib_i2cs_sub_state		;1
		dec	lib_i2cs_bit_count		;1
		sz				;1
		retp				;3 = 15 + 7 = 22
		inc	lib_i2cs_state		;1
		retp				;3 = 16 + 7 = 23
						; Stay in this i2cs_state until all 8 bits
						; are received	
_i2cs_send_ack
:state1
		call	i2cs_clrbSDA			;Clear SDA to set up ACK
		inc	lib_i2cs_sub_state		;1
		retp				;3 = 12 + 7 = 19
:state2					
		call	i2cs_setbSCL			;Allow the clock to go high
		sb	lib_i2cs_pres_scl		;1 ;Check to see if the clock is high.
		retp				;3 = 12 + 7 = 19
		inc	lib_i2cs_sub_state		;1
		retp				;3 = 13 + 7 = 20
:state3
		snb	lib_i2cs_pres_scl		;1
		retp				;3 = 11 + 7 = 18
		call	i2cs_clrbSCL
		call	i2cs_setbSDA
		clr	lib_i2cs_sub_state		;1
		inc	lib_i2cs_state		;1
		retp				;3 = 15 + 7 = 22

_i2cs_write_byte
:state1
		rl	lib_i2cs_byte		;1
		call	i2cs_setbSDA
		sc				;1
		call	i2cs_clrbSDA
		inc	lib_i2cs_sub_state		;1
		retp				;3 = 15 + 7 = 22
:state2
		call	i2cs_setbSCL			;allow scl to go high, indicating slave is
						; ready
		snb	lib_i2cs_pres_scl		;1 ;If SCL is not yet high, don't proceed to next state.
		inc	lib_i2cs_sub_state		;1
		retp				;3 = 13 + 7 = 20
:state3
		snb	lib_i2cs_pres_scl		;1
		retp				;3 = 11 + 7 = 18
		clr	lib_i2cs_sub_state		;1
		call	i2cs_clrbSCL			;Put scl low again to indicate slave is processing
		call	i2cs_setbSDA			;release sda line again
		dec	lib_i2cs_bit_count		;1
		sz				;1 ;once we've done all 8 bits, go to next state	
		retp				;3 = 16 + 7 = 23
		inc	lib_i2cs_state		;1
		retp				;3 = 17 + 7 = 24

	;*********************************************************************************
	; State: i2cs_idle
	;	i2cs_idle is the state used when no start condition has occurred
	;	and the device has not been addressed.  The I2C slave will sit in this
	;	state until the state is changed by the independant subroutine 
	;	i2cs_get_start_stop.  
	;*********************************************************************************
i2cs_idle	call	i2cs_setbSCL
		jmp	i2cs_setbSDA

	;*********************************************************************************
	; State: i2cs_wait_for_scl_low
	;	This state simply waits until SCL goes low.  It is entered once
	;	i2cs_get_start_stop encounters a start condition.  Once SCL goes low,
	;	the state is incremented
	;*********************************************************************************
i2cs_wait_for_scl_low
		snb	lib_i2cs_pres_scl		;1
		retp				;3 = 4 + 7 = 11
		call	i2cs_clrbSCL
		inc	lib_i2cs_state		;1
		retp				;3 = 6 + 7 = 13
	;*********************************************************************************
	; State: i2cs_get_address
	;	This state prepares the i2cs_read_byte routine to get 8-bits of data. 
	;	It increments the state to i2cs_read_byte before exiting.
	;*********************************************************************************
i2cs_get_address
		mov	w,#8			;1
		mov	lib_i2cs_bit_count,w	;1
		inc	lib_i2cs_state		;1
		retp				;3 = 6 + 7 = 13
	;*********************************************************************************
	; State: i2cs_process_address
	;	This state simply performs a quick calculation to figure out if this slave
	;	was just addressed.  If it was not, then it changes the state back to 
	;	i2cs_idle.  If this slave was addressed, then this state prepares to either
	;	read data or write data, depending on the bit 1 of the first byte received.
	;*********************************************************************************
i2cs_process_address
		mov	w,lib_i2cs_byte		;1 ;Get the received I2C byte
		xor	w,lib_i2cs_address		;1 ;Compare with the address of this slave
		and	w,#$FE			;1 ;And out the R/!W bit
		snz				;1
		jmp	:slave_addressed	;3 ;If it is equal, send an ACK
		clr	lib_i2cs_state		;1 ;If the device was not addressed, then
		clr	lib_i2cs_sub_state		;1 ;go back to the idle state
		retp				;3 = 12 + 7 = 19
:slave_addressed
		call	i2cs_clrbSCL	
						; If bit zero of the incoming byte is a 0,
		snb	lib_i2cs_byte.0		;1 ;then go to the being written state, otherwise
		jmp	:being_read		;3 ;go to the being read state
		mov	w,#(i2cs_being_written-i2cs_idle_loc) ;1
		mov	lib_i2cs_state,w		;1
		retp				;3 = 14 + 7 = 21

:being_read	mov	w,#(i2cs_being_read-i2cs_idle_loc);1
		mov	lib_i2cs_state,w		;1
		retp				;3 = 18 + 7 = 25
	;*********************************************************************************
	; State: i2cs_read_data
	;	This state prepares i2cs_read_byte to receive 8 bits of data.  It increments
	;	the state before exiting.
	;*********************************************************************************
i2cs_read_data
		mov	w,#8			;1
		mov	lib_i2cs_bit_count,w	;1
		inc	lib_i2cs_state		;1
		retp				;3 = 13 + 7 = 20

	; State: i2cs_read_again
	;	This state prepares to re-enter i2cs_read_data to receive another 8 bits of data.
	;	It stays in this state until i2cs_rx_flag is cleared.
	;---------------------------------------------------------------------------------
i2cs_read_again
		snb	lib_i2cs_rx_flag		;wait till mainline code has grabbed data
		retp
		setb	lib_i2cs_written_again	; identifies 2nd or later databyte, cleared by stop condition
		mov	w,#(i2cs_being_written_again-i2cs_idle_loc) ;1
		mov	lib_i2cs_state,w		;  ;we may receive another byte
		retp				;3 = 13 + 7 = 20

	;*********************************************************************************
	; State: i2cs_process_data
	;	This state processes a byte of data which was just received.  It moves the 
	;	byte that was just received into the i2cs_data_in register, and sets the 
	;	i2cs_event_flag to indicate an i2cs event and sets the i2cs_rx_flag to 
	;	indicate that a byte of data was received by the slave.
	;*********************************************************************************
i2cs_process_data
		setb	lib_i2cs_rx_flag		;1
		inc	lib_i2cs_state		;1
		retp				;3 = 15 + 7 = 22

	;*********************************************************************************
	; State: i2cs_send_data
	;	This state prepares the i2cs_write_byte state to send the 8-bits of data
	;	in the i2cs_data_out register.
	;*********************************************************************************
i2cs_send_data
		inc	lib_i2cs_state		;1
		clr	lib_i2cs_sub_state		;1
		mov	w,#8			;1
		mov	lib_i2cs_bit_count,w	;1
		retp				;3 = 17 + 7 = 24

_i2cs_get_ack
:state1		call	i2cs_setbSCL
		sb	lib_i2cs_pres_scl		;1
		retp				;3 = 12 + 7 = 19
		snb	lib_i2cs_pres_sda		;1
		jmp	:NACK			;3 = 13 + 7 = 20
		inc	lib_i2cs_sub_state		;1
		retp				;3 = 14 + 7 = 21

:NACK		clr	lib_i2cs_sub_state		;1
		clr	lib_i2cs_state		;1
		retp				;3 = 18 + 7 = 25

:state2		snb	lib_i2cs_pres_scl		;1
		retp				;3 = 11 + 7 = 18
		call	i2cs_clrbSCL
		clr	lib_i2cs_sub_state		;1
		inc	lib_i2cs_state		;1
		retp				;3 = 14 + 7 = 21

	; State: i2cs_request_for_data
	;	This state tells the mainline code to supply data in the i2cs_byte register.
	;---------------------------------------------------------------------------------
i2cs_request_for_data
		setb	lib_i2cs_data_needed	;1 mainline code must fill i2cs_byte and clear i2cs_data_needed
		inc	lib_i2cs_state
		retp				;3 = 13 + 7 = 20

	;*********************************************************************************
	; State: i2cs_waiting_for_data
	;	This state waits for the mainline routine to put some valid data in the
	;	I2C_data_out register.  This state is only reached if the master has 
	;	indicated it wants more data by sending an ACK after the last byte.
	;*********************************************************************************
i2cs_waiting_for_data
		snb	lib_i2cs_data_needed	;1
		retp				;3 = 13 + 7 = 20
		inc	lib_i2cs_state		;1
		retp				;3 = 16 + 7 = 23 

;*****************************************************************************************
; I2C Slave subroutines
;*****************************************************************************************

;*****************************************************************************************
; End of I2C Slave state machine
;*****************************************************************************************	
noexpand
code		=	$
if _lib_i2cSlave_highsize > code - begin
  error P2W "_i2cSlave_highsize too large"
endif
if _lib_i2cSlave_highsize < code - begin
  error "_i2cSlave_highsize too small"
endif
endm

;end of file lib_i2cSlave.inc
