;file lib_uartReceive.inc

_lib_uart_rx_var	macro	nr, prefix
			org	data
expand
vp??nr??_bank		equ	($ & $F8) | nr
lib_rx_count		=	$+0		;number of bits
lib_rx_divide		=	$+1		;timing counter
lib_rx_low		=	$+2		;shift register
lib_rx_high		=	$+3		;value
lib_rx_format		=	$+4
  lib_rx_bits_0		=	lib_rx_format.0	;bits 1 and 0 specify number of total bits
  lib_rx_bits_1		=	lib_rx_format.1	;01 -> 9bits, 10 -> 10bits, 11 -> 11bits (start+data+parity+stop)
  lib_rx_7bits		=	lib_rx_format.2	;set for 7 databits (0 = 8 databits)
  lib_rx_even		=	lib_rx_format.3	;set for even parity
  lib_rx_odd		=	lib_rx_format.4	;set for odd parity
  lib_rx_data		=	lib_rx_format.5	;set if valid value, cleared after value grabbed
  lib_rx_enable		=	lib_rx_format.6	;set to enable uart
  lib_rx_parity		=	lib_rx_format.7	;parity bit placeholder
lib_rx_baud		=	$+5		;baudrate
lib_rx_datapin		=	$+6		;high nib = port (5 to 9), low nib = pin (0 to 7)
  lib_rx_error		=	lib_rx_datapin.3	;set if parity error
lib_rx_hspin		=	$+7		;high nib = port (5 to 9), low nib = pin (0 to 7)
prefix??_??bank		equ	($ & $F0) | nr
prefix??_??count	ds	1		;number of bits
prefix??_??divide	ds	1		;timing counter
prefix??_??low		ds	1		;shift register
prefix??_??high	ds	1		;value
prefix??_??format	ds	1
  prefix??_??bits_0	equ	prefix??_??format.0	;bits 1 and 0 specify number of total bits
  prefix??_??bits_1	equ	prefix??_??format.1	;01 -> 9bits, 10 -> 10bits, 11 -> 11bits (start+data+parity+stop)
  prefix??_??7bits	equ	prefix??_??format.2	;set for 7 databits (0 = 8 databits)
  prefix??_??even	equ	prefix??_??format.3	;set for even parity
  prefix??_??odd	equ	prefix??_??format.4	;set for odd parity
  prefix??_??data	equ	prefix??_??format.5	;set if valid value, cleared after value grabbed
  prefix??_??enable	equ	prefix??_??format.6	;set to enable uart
  prefix??_??parity	equ	prefix??_??format.7	;parity bit placeholder
prefix??_??baud	ds	1		;baudrate
prefix??_??datapin	ds	1		;high nib = port (5 to 9), low nib = pin (0 to 7)
  prefix??_??error	equ	prefix??_??datapin.3	;set if parity error
prefix??_??hspin	ds	1		;high nib = port (5 to 9), low nib = pin (0 to 7)
noexpand
data			=	$
endm

lib_uartReceive		macro	nr, prefix, datapin, hspin, format, baud, thread
if nr >= maxvp
  error "vp number too high"
endif
ifndef vp_uartReceive
vp_uartReceive
endif
ifndef vp_dirbuf
vp_dirbuf
endif
ifndef vp_parity
vp_parity
endif
vp??nr??_thread	= thread
_lib_uart_rx_var nr,prefix
			org	code
expand
vp??nr??_init
			mov	fsr,#vp??nr??_bank
			mov	lib_rx_format,#format
			mov	lib_rx_baud,#baud
			mov	lib_rx_datapin,#datapin
			mov	lib_rx_hspin,#hspin
noexpand
code			=	$
expand
			org	vp??nr??_isr
			jmp	@uart_rx_isr
noexpand
endm

_lib_uartReceive_lowsize	equ	$002
_lib_uartReceive_lowhalf	macro
			org	code
	begin	=	$
expand
uart_rx_isr		jmp	_uart_rx_isr
uart_rx_task		jmp	_uart_rx_task
noexpand
code		=	$
if _lib_uartReceive_lowsize > code - begin
  error P2W "_uartReceive_lowsize too large"
endif
if _lib_uartReceive_lowsize < code - begin
  error "_uartReceive_lowsize too small"
endif
endm

_lib_uartReceive_highsize	equ	$05D
_lib_uartReceive_highhalf	macro
			org	code
	begin	=	$
expand
_uart_rx_isr
			mov	fsr,vp_bank		;select target uart rambank
			jnb	lib_rx_enable,:rxdone
			mov	w,<>lib_rx_datapin		;port (ra to re) in lownib
			or	w,#$F0
			mov	fsr,w			;portA_dirbuf to portE_dirbuf
			sub	fsr,#portA_dirbuf-5	;ra to re
			mov	w,indf			;read all pins
			mov	vp_portpin,w		;save pin states
			mov	fsr,vp_bank
			mov	w,lib_rx_datapin	;pin number 0 to 7 in lownib
			call	pin_mask			;get mask for pin
			and	vp_portpin,w		;extract bit for pin
			test	lib_rx_count		;currently receiving byte?
			sz				;no
			jmp	:rx1			;yes, jump ahead
			mov	w,--lib_rx_format		;ready bits for start
			and	w,#$03			;number of bits 1, 2 or 3
			or	w,#$08			;9, 10 or 11
			test	vp_portpin		;start bit?
			snz				;no
			mov	lib_rx_count,w		;yes, renew bit count
			mov	w,>>lib_rx_baud		;half bittime delay
			add	w,lib_rx_baud		;plus bittime delay -> start delay
			mov	lib_rx_divide,w
:rx1
			clc
			test	vp_portpin
			sz
			stc
			decsz	lib_rx_divide		;middle of next bit?
          		jmp	:rxdone			;no
			dec	lib_rx_count		;last bit?
			jz	:rx2			;yes
			rr	lib_rx_low			;no, rotate in new bit
			movb	lib_rx_parity,C		;save b0
			jmp	:rx3			;renew delay time
:rx2
			mov	w,lib_rx_low		;save byte to allow receive another byte
			mov	lib_rx_high,w
			setb	lib_rx_data			;mark byte received
:rx3
			mov	w,lib_rx_baud		;renew delay time
			mov	lib_rx_divide,w
:rxdone
			retp
_uart_rx_task
			sb	lib_rx_data
			retp
			movb	C,lib_rx_parity		;get last saved b0
			jb	lib_rx_format.1,:uit1
			clc				;adjust for 7 databits + no parity (total bits is 9)
			rr	lib_rx_high
			rr	lib_rx_high
:uit1
			snb	lib_rx_format.0
			rl	lib_rx_high			;adjust for 8 databits + parity (total bits is 11)
			jnb	lib_rx_7bits,:uit2
			movb	C,lib_rx_high.7		;save possible parity in case 7 databits
			clrb	lib_rx_high.7		;clear b7 in case 7 databits
:uit2
			movb	lib_rx_parity,C		;save possible parity bit
			clrb	lib_rx_error		;assume no error
			jb	lib_rx_odd,:uit3
			jb	lib_rx_even,:uit3
			jmp	:uit4
:uit3
			mov	w,lib_rx_high
			call	@even_parity
			snb	lib_rx_odd
			not	vp_portpin			;invert calculated even parity bit for odd parity
			addb	vp_portpin,lib_rx_parity		;if parity matches, b0 will be cleared
			snb	vp_portpin.0
			setb	lib_rx_error		;parity error
:uit4
			clrb	lib_rx_data			;mark byte grabbed
			mov	w,fsr			;isr_flag to use in lownib
			call	pin_mask
			or	vp_flags,w		;C isr() gets called
			retp
noexpand
code			=	$
if _lib_uartReceive_highsize > code - begin
  error P2W "_uartReceive_highsize too large"
endif
if _lib_uartReceive_highsize < code - begin
  error "_uartReceive_highsize too small"
endif
endm

;end of file lib_uartReceive.inc
